//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Baud generator
//
module baud_generator (
  clk_i, rst_i, divisor_msb_i, divisor_lsb_i, load_divisor_i,
  baudout_o  
  
  );
 
  input clk_i;                          // Input clock signal
  input rst_i;                          // Input reset signal
  input [7:0] divisor_msb_i;            // MSB of the divisor latch
  input [7:0] divisor_lsb_i;            // LSB of the divisor latch
  input load_divisor_i;                 // Input signal to load divisor registers
  output baudout_o;                     // Output baud rate (16 times Tx clock)
  
  reg baudout_o = 1'b0;               // Initial value is required
  reg [15:0] divisor_reg;            // Register to store divisor latch 
  
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          baudout_o <= ~baudout_o;      // Make it oscillating while reset to set circuits depending
                                        // on edges of baudout  
          divisor_reg <= {divisor_msb_i,divisor_lsb_i};
        end
      else
        begin
          case (load_divisor_i)
            1'b0:
              begin
                if (divisor_reg ==  16'h1)
                  begin
                    baudout_o <= !baudout_o;
                    divisor_reg <= {divisor_msb_i,divisor_lsb_i};
                  end
                else
                  divisor_reg <= divisor_reg - 1'b1;
              end
            1'b1:                 // Normally asserted when new value to the divisor is loaded
              divisor_reg <= {divisor_msb_i,divisor_lsb_i};
          endcase
        end  
    end
endmodule
