//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Bypass data from stages M and A to the stage E
//
module bypass (
 clk_i, rst_i, rsae_i, rtae_i, rdam_i, rdaa_i, rse_i, rte_i,
 rdm_i, rda_i, rdl_aligned_i, rdlaa_i, rtypem_i, rtypea_i,
 loada_i,
 rs_o, rt_o
  );
  input clk_i;                // Input clock
  input rst_i;                // Input reset signal
  input [4:0] rsae_i;         // rs register address from the E stage
  input [4:0] rtae_i;         // rt register address from the E stage
  input [4:0] rdam_i;         // rd register address from the M stage
  input [4:0] rdaa_i;         // rd register address from the A stage
  input [31:0] rse_i;         // rs register from the E stage   
  input [31:0] rte_i;         // rt register from the E stage
  input [31:0] rdm_i;         // rd register from the M stage
  input [31:0] rda_i;         // rd register from the A stage
  input [31:0] rdl_aligned_i; // Aligned input from Load operation in A stage
  input [4:0] rdlaa_i;         // Address of loaded register from A stage
  input rtypem_i;             // Instruction of R-type flag in M stage
  input rtypea_i;             // Instruction of R-type flag in A stage 
  input loada_i;              // Instruction of load type in the A stage
  output [31:0] rs_o;         // rs register for ALU
  output [31:0] rt_o;         // rt register for ALU


  reg [31:0] rs_o;
  reg [31:0] rt_o;
  
// Two muxes to specify input for the ALU
  always @ (*)
    begin
      if ({1'b1,rsae_i} === {rtypem_i,rdam_i})
        rs_o = rdm_i;
      else if (rtypea_i && (rsae_i === rdaa_i))
        rs_o = rda_i;
      else if (loada_i && (rsae_i === rdlaa_i))
        rs_o = rdl_aligned_i;
      else
        rs_o = rse_i;
    end
    
  always @ (*)
    begin
      if ({1'b1,rtae_i} === {rtypem_i,rdam_i})
        rt_o = rdm_i;
      else if (rtypea_i && (rtae_i === rdaa_i))
        rt_o = rda_i;
      else if (loada_i && (rtae_i === rdlaa_i))
        rt_o = rdl_aligned_i;
      else
        rt_o = rte_i;
    end
 

endmodule
