/* elf2bin.c */

/* This program converts elf file into bin image
It has been made from the elf_viewer so some redundant logic 
presents here.
Extraction of the .text section done in the print_sht function */

#undef DEBUG /* Compile all parts of the program intended for debugging */
//#define DEBUG
//#define DEBUG2
#include "header.h"
#include <stdio.h>
#include <stdlib.h>

/* Prototypes of functions */
 void if_elf(Elf32_Ehdr* );
 Elf32_Shdr* shdr_dyn_memory(Elf32_Ehdr*);
void pop_sh(Elf32_Ehdr*, Elf32_Shdr*, FILE* fp);
void print_shdr_entry(Elf32_Shdr*, int);
char* str_dyn_memory(Elf32_Ehdr*, Elf32_Shdr*);
void pop_str_t(Elf32_Ehdr*, Elf32_Shdr*, char*, FILE*);
void print_str_t(Elf32_Ehdr*, Elf32_Shdr*, char*);
void print_sht(Elf32_Ehdr*, Elf32_Shdr*, char*, univ_struct*, FILE*);
void elf_hdr_print(Elf32_Ehdr*);
int if_sym_tab(Elf32_Ehdr*, Elf32_Shdr*, Elf32_Word);
Elf32_Sym* symtab_dyn_memory(Elf32_Shdr*, int);
void sym_tab_print(Elf32_Shdr*, Elf32_Sym*, FILE*, int, char*);
void print_symtab_hex(Elf32_Shdr*, Elf32_Sym*, int);
void pop_sym_t(Elf32_Shdr*, Elf32_Sym*, FILE*, int);
Elf32_Phdr* phdr_dyn_memory(Elf32_Ehdr*);
void pop_pht(Elf32_Ehdr*, Elf32_Phdr*, FILE*);
void print_pht(Elf32_Ehdr*, Elf32_Phdr*);
int if_dyn_pht(Elf32_Ehdr*, Elf32_Phdr*);
Elf32_Dyn* dyn_dyn_memory(Elf32_Phdr* ,int);
void pop_dynamic(Elf32_Dyn*, Elf32_Phdr*, FILE*, int);
void print_dynamic(Elf32_Phdr*, Elf32_Dyn*, FILE*, univ_struct*);
void print_dynamic_hex(Elf32_Phdr*, Elf32_Dyn*);
 

int main (int argc, char* argv[])
{
Elf32_Ehdr elf_header; /* Elf header structure. It has one member, not array */
Elf32_Phdr* p_header_ptr=0; /* Pointer to array of the Program Header Table sturctures */
Elf32_Shdr* s_header_ptr=0; /* Pointer to array Section Header Table structures */
Elf32_Sym* symtab_ptr=0; /* Pointer to Symbol Table */
Elf32_Sym* dynsym_ptr=0; /* Pointer to Dynsym Table */
Elf32_Dyn* dyn_ptr=0; /* Pointer to DYNAMIC segment's array */
char* string_t_ptr=0; /* Pointer to the String Table */
int temp=0; /* For universal purposes */
/* Define structure for universal purposes */
univ_struct struct1;
struct1.dynstr_offset=0;
/* Check if the program invoked properly */

if (argc!=2)
 {
  printf("Usage: %s <file_name>\n",argv[0]);
  exit(1);
  }
/* Open the file */
FILE* fp;
if((fp = fopen (argv[1], "rb")) == NULL)
 {
  perror("fopen() failed");
  exit(1);
 }

/* Read data into the structure elf_header */
size_t count;
if (( count = fread(&elf_header,sizeof(Elf32_Ehdr),1, fp)) != 1)
 {
    if (!feof(fp))
     {
  	printf("Error while reading file %s in main()\n",argv[1]);
  	fclose(fp);
  	exit(1);
      }	
 }
/* After the data is read into elf_header structure
 we need to check if the file is ELF format at all
*/
if_elf(&elf_header);

/* Print some data from ELF header in readable format */
elf_hdr_print(&elf_header);
/* We can do manipulations with Program Header Table
only if it exists
*/
if((elf_header.e_phoff)==0)
	printf("The file %s does not have Program Header Table\n",argv[1]);
else {
/* Allocate memory for PHT */
p_header_ptr=phdr_dyn_memory(&elf_header);
/* Populate PHT */
pop_pht(&elf_header, p_header_ptr,fp);
/* Print PHT in readable form */
print_pht(&elf_header, p_header_ptr);

/* Check if PHT has Dynamic Section. If so we need to print
it's contents */
   if ((temp=if_dyn_pht(&elf_header, p_header_ptr))==0xffffffff)
	printf("The file does not have DYNAMIC section\n");
   else{
/* Allocate memory for DYNAMIC section */
dyn_ptr=dyn_dyn_memory(p_header_ptr ,temp);
pop_dynamic(dyn_ptr, p_header_ptr, fp, temp);
#ifdef DEBUG2
print_dynamic_hex(p_header_ptr, dyn_ptr);
#endif
   }/*End of else for PHT Dynamic Section */
} /* End of else for Program Header Table manipulations */

/* We can do manipulations with Section Header Table
only if it exists
*/
if((elf_header.e_shoff)==0)
   {
   	printf("The file %s does not have Section Header Table\n",argv[1]);
   }
   else
{
/* Allocate memory for SHT */
s_header_ptr=shdr_dyn_memory(&elf_header);
/* Populate Section header table */
pop_sh(&elf_header,s_header_ptr,fp);
/* Dynamically allocate memory for String Table */
string_t_ptr=str_dyn_memory(&elf_header,s_header_ptr);

/* Populate String Table */  
pop_str_t(&elf_header, s_header_ptr, string_t_ptr, fp);

/* Print content of the Section Header Table in readable format */
print_sht(&elf_header,s_header_ptr,string_t_ptr, &struct1, fp);
/* The function checks if there is a Symbol Table
in the file and if so memory is allocated for it 
and its contents is printed */

if(!(temp = if_sym_tab(&elf_header,s_header_ptr,SHT_SYMTAB)))
	printf("\nThe file doesn't have Symbol Table\n");
else {
/* Allocate memory for string table */
symtab_ptr = symtab_dyn_memory(s_header_ptr,temp);
/* Populate symbol table */
pop_sym_t(s_header_ptr, symtab_ptr, fp, temp);
#ifdef DEBUG2
print_symtab_hex(s_header_ptr, symtab_ptr, temp);
#endif
/* Prints Symbol Table in readable form */
#ifdef DEBUG2
printf("String Table for Section Header Table has index %i\n",elf_header.e_shstrndx);
#endif
sym_tab_print(s_header_ptr, symtab_ptr, fp, temp, string_t_ptr);
} /* end of SYMTAB else */
/* Similar to Symbol Table. Check if the file has DYNSYM 
table, and if so allocate memory and print its content
*/
if(!(temp = if_sym_tab(&elf_header,s_header_ptr,SHT_DYNSYM)))
	printf("\n The file doesn't have DYNSYM table\n");
else {
/* Allocate memory for dynsym table */
dynsym_ptr = symtab_dyn_memory(s_header_ptr,temp);
/* Populate dynsym table */
pop_sym_t(s_header_ptr, dynsym_ptr, fp, temp);
/* Print DYNSYM table in readable form */
printf("\033[32m\n%25s\033[0m","DYNSYM TABLE");
sym_tab_print(s_header_ptr, dynsym_ptr, fp, temp, string_t_ptr);
print_dynamic(p_header_ptr, dyn_ptr, fp, &struct1);
}/* end of DYNSYN else */
}/* end of Section Header Table else */
/* Free dinamically allocated memory
Only if the file has SHT and as a result memory has been
allocated*/
if(s_header_ptr || string_t_ptr)
{
#ifdef DEBUG2
printf("Memory from SHT and String Table is being freed\n");
#endif
free(s_header_ptr);
free(string_t_ptr);
}
free(p_header_ptr);
free(dynsym_ptr);
free(dyn_ptr);
/* After everything is done file descriptor must be closed */
if (fclose(fp))
	perror("fclose() failed");
return 0;

 } /* This is closing bracket for the main() function */

/*===========================================================
                FUNCTIONS DEFINITIONS
=============================================================*/
 void if_elf(Elf32_Ehdr* header)
{
/* Check if this is ELF file at all */
if(header->e_ident[1] !='E')
	{ printf("This file is not ELF format\n");
	  exit(1);
	}
if(header->e_ident[2] !='L')
	{ printf("This file is not ELF format\n");
	  exit(1);
	}
if(header->e_ident[3] !='F')
	{ printf("This file is not ELF format\n");
	  exit(1);
	}
}

/* Function to populate section header table array */
void pop_sh(Elf32_Ehdr* elf_header,Elf32_Shdr* sh_header, FILE* fp)
{

/* Position pointer of the opened file to the beginning of SHT */
if((fseek(fp,elf_header->e_shoff,SEEK_SET)))
  	{
  	   perror("fseek() failed");
  	   exit(1);
  	}

/* Cycle to read SHT from file into array of structures */
   	
int count=0; 
if ((count = fread(sh_header,elf_header->e_shentsize,elf_header->e_shnum,fp)) !=elf_header->e_shnum)
   {
    printf("Error while reading file str_table_parse(2)\n");
    fclose(fp);
    exit(1);
   }
#ifdef DEBUG
printf("%i members in SHT\n",count);
#endif
}

  /* Function to allocate dynamically memory for the SHT */
  Elf32_Shdr* shdr_dyn_memory(Elf32_Ehdr* elf_header)
  {
    Elf32_Shdr* sh_table_ptr;
  
if(sizeof(Elf32_Shdr)!=elf_header->e_shentsize)
   {
   	printf("sizeof(Elf32_Shdr)!=elf_header.e_shentsize\n");
   	printf("Structure of Section Header Table entry is unknown\n");
	exit(1);
   }
if ((sh_table_ptr= malloc(elf_header->e_shnum*elf_header->e_shentsize))==NULL)
  {
    perror("malloc() failed");
    exit(1);
  }
  return sh_table_ptr;
}


/* This function primarily used for test purposes only
It's to print hex values of the SHT members
One parameter is pointer to array of structures Elf32_Shdr and 
another parameter is which structure to print*/
void print_shdr_entry(Elf32_Shdr* sh_table, int num)
{
 Elf32_Shdr* table=sh_table+num;
  printf("\033[32m %#010x  %#010x  %#010x  %#010x  %#010x  %#010x  %#010x  %#010x  %#010x  %#010x  \033[0m\n",table->sh_name,table->sh_type,table->sh_flags,table->sh_addr,table->sh_offset,table->sh_size,table->sh_link,table->sh_info,table->sh_addralign,table->sh_entsize);
 }

/* Function to dynamically allocate memory for the String Table */
char* str_dyn_memory(Elf32_Ehdr* elf_hdr_ptr,Elf32_Shdr* sht_hdr_ptr)
{
char* p;
/* Find out which entry in the SHT is for String Table and align
pointer accordingly */
sht_hdr_ptr=sht_hdr_ptr+(elf_hdr_ptr->e_shstrndx);
#ifdef DEBUG
printf("String Table size:%#010x\n",sht_hdr_ptr->sh_size);
#endif
if ((p=malloc(sht_hdr_ptr->sh_size))==NULL)
  {
    perror("malloc() failed");
    exit(1);
  }
return p;

}
/* This function is to populate String Table */
void pop_str_t(Elf32_Ehdr* elf_hdr_ptr, Elf32_Shdr* sht_hdr_ptr, char* str_t_ptr,FILE* fp)
{
/* Align file pointer to the beginning of the String Table */
sht_hdr_ptr=sht_hdr_ptr+(elf_hdr_ptr->e_shstrndx);

if ((fseek(fp,sht_hdr_ptr->sh_offset,SEEK_SET)))
 {
  perror("fseek() failed");
  exit(1);
  }

int count=0;
if ((count = fread(str_t_ptr,1,sht_hdr_ptr->sh_size,fp)) !=sht_hdr_ptr->sh_size)
   {
    printf("Error while reading file in pop_str_t\n");
    fclose(fp);
    exit(1);
  }
#ifdef DEBUG
printf("%#010x bytes read into String Table\n",count);
#endif
}
/* This function is primarily for debug purposes
 It prints content of the String Table */
 void print_str_t(Elf32_Ehdr* elf_hdr_ptr, Elf32_Shdr* sht_hdr_ptr, char* str_t_ptr)
 {
 /* Define size of String Table */
 sht_hdr_ptr=sht_hdr_ptr+(elf_hdr_ptr->e_shstrndx);
 printf("\nprint_str_t() Size of sting table is %#010x\n",sht_hdr_ptr->sh_size);
 /* Print String Table */
 int i=0;
 for(i=0;i<sht_hdr_ptr->sh_size;i++)
    {
       if(str_t_ptr[i]==0)
          printf("\n");
       else
           printf("%c",str_t_ptr[i]);
    }
}
/* Function to print Section Header Table entries in readable form */
void print_sht(Elf32_Ehdr* elf_hdr_ptr, Elf32_Shdr* sht_ptr, char* str_t_ptr, univ_struct* struct_ptr, FILE*fp)
{
FILE* fpb;
unsigned char* buf_ptr;
char* control_sign;
*control_sign = '7';

if ((fpb = fopen ("mips.bin","w")) == NULL)
  {
    perror("fopen() failed");
    exit(1);
  }
/* Print header for displaying, similar to the one from objdump */
printf("\n\033[32m%15s\033[0m","Sections:");
printf("\nIdx Name\t      Size\tVMA\t\tFile off\tAlgn\n");

/* By the time this function is called Section Header Table and
String table already read from file and reside in arrays pointed
by sht_ptr and str_t_ptr accordingly

Section Header Table's Index 0 contains isn' interesing for parsing
as it contains no meaningful data.

Start parsing from Index 1*/
sht_ptr++;
#ifdef DEBUG2
printf("Section Header Table has %i entries\n",elf_hdr_ptr->e_shnum);
#endif
int i=0;
int idx=0;
char* init_ptr = str_t_ptr;
for(i=1; i < elf_hdr_ptr->e_shnum;i++)
   {
     printf("%i ",i);
     idx=(sht_ptr->sh_name);
     str_t_ptr +=idx;
     if (!strcmp(str_t_ptr,".dynstr"))
     struct_ptr->dynstr_offset=sht_ptr->sh_offset;
     printf("%-18s",str_t_ptr);
     printf("%08x\t",sht_ptr->sh_size);
     printf("%08x\t",sht_ptr->sh_addr);
     printf("%08x\t",sht_ptr->sh_offset);
     printf("%x\t",sht_ptr->sh_addralign);
     printf("\n\t\t\t");
    
    if (!strcmp(str_t_ptr,".text"))
    {	// Text section. Need to read data from the section
      if (fseek(fp,sht_ptr->sh_offset,SEEK_SET))
        {
          perror("fseek() failed");
          exit(1);
        }
      if (( buf_ptr = malloc(sht_ptr->sh_size)) == NULL)
        {
          perror("malloc() failed");
          exit(1);
        }
      int count = 0;
      if ((count = fread(buf_ptr,sizeof(unsigned char),sht_ptr->sh_size,fp)) != sht_ptr->sh_size)
        {
          printf("Error while reading .text section\n");
          fclose(fp);
          fclose(fpb);
          exit(1);
        } 
      else
        {
          printf("%i bytes read from the .text section\n",count);
        }
     int  count1 = 0;
      if ((count1 = fwrite(buf_ptr,sizeof(unsigned char),sht_ptr->sh_size,fpb)) != sht_ptr->sh_size)
        {
          printf("Error while writing binary data from .text section\n");
          printf("%i bytes has been written\n",count1);
          fclose(fp);
          fclose(fpb);
          exit(1);
        }
      else
       {
         printf("%i bytes written to the mips.bin file\n",count1);
       }
     count1 = 0;
     if ((count1 = fwrite(control_sign,1,1,fpb)) != 1)    // Need to write control byte at the end
       {
         printf("Error while writing control byte\n");
         fclose(fp);
         fclose(fpb);
         exit(1);
       }
     else
       printf("Control byte has been successfully written\n");

    }
/* Type of section in readable form */
printf("Type: ");
switch (sht_ptr->sh_type) {
	case SHT_NULL:
	   printf("NULL\t");
	   break;
	case SHT_PROGBITS:
	   printf("PROGBITS\t");
	   break;
	case SHT_SYMTAB:
	   printf("SYMTAB\t");
	   break;
	case SHT_STRTAB:
	   printf("STRTAB\t");
	   break;
	case SHT_RELA:
	   printf("RELA\t");
	   break;
	case SHT_HASH:
	   printf("HASH\t");
	   break;
	case SHT_DYNAMIC:
	   printf("DYNAMIC\t");
	   break;
	case SHT_NOTE:
	   printf("NOTE\t");
	   break;
	case SHT_NOBITS:
	   printf("NOBITS\t");
	   break;
	case SHT_REL:
	   printf("REL\t");
	   break;
	case SHT_SHLIB:
	   printf("SHLIB\t");
	   break;
	case SHT_DYNSYM:
	   printf("DYNSYM\t");
	   break;
	default:
	   printf("Reserved type\t");
	}   
/* Analyze flags */
printf("Flags: ");
int flags=sht_ptr->sh_flags;
if(flags&SHF_WRITE)
	printf("WRITE\t");
if(flags&SHF_ALLOC)
	printf("ALLOC\t");
if(flags&SHF_EXECINSTR)
	printf("EXECINSTR\t");
	
	printf("\n");   
     sht_ptr++;
     str_t_ptr=init_ptr;
    }
if (fclose(fpb))
  {
    perror("fclose() failed");
    exit(1);
  }
}
/* Function to analyze Elf header and print some 
readable information */
void elf_hdr_print(Elf32_Ehdr* elf_header)
{
/* Analyze type of file */
printf("\nType: ");
switch (elf_header->e_type) {
	case ET_NONE:
		printf("NONE\t");
		break;
	case ET_REL:
	 	printf("REL\t");
		break;
	case ET_EXEC:
		printf("EXEC\t");
		break;
	case ET_DYN:
		printf("DYN\t");
		break;
	case ET_CORE:
	 	printf("CORE\t");
		break;
	default:
		printf("Unknown type\t");
}
/* Analyze type of machine */
printf("Arch: ");
switch (elf_header->e_machine) {
	case EM_NONE:
		printf("No machine\t");
		break;
	case EM_M32:
		printf("AT&T WE 32100\t");
		break;
	case EM_SPARC:
		printf("SPARC\t");
		break;
	case EM_386:
		printf("Intel 80386\t");
		break;
	case EM_68K:
		printf("Motorola 68000\t");
		break;
	case EM_88K:
		printf("Motorola 88000\t");
		break;
	case EM_860:
		printf("Intel 80860\t");
		break;
	case EM_MIPS:
		printf("MIPS RS3000\t");
		break;
	default:
		printf("Unknown architecture\t");
	}
/* Analyze version */
printf("Version: ");
switch (elf_header->e_version) {
	case EV_NONE:
		printf("Invalid version\t");
		break;
	case EV_CURRENT:
		printf("Current version\t");
		break;
	default:
		printf("Unknown version\t");
	}
	
/* Entry point */
printf("\nEntry point: %08x\n",elf_header->e_entry);
}

/* The function is to find out if there is a 
symbol table (SYMTAB) or dynamic linking symbol table 
(DYNSYM) in the file. And if there is, it returns
the index to it in the Section Header Table. Type of table
to find is passed as argument
*/
int if_sym_tab(Elf32_Ehdr* elf_hdr_ptr,Elf32_Shdr* sht_ptr, Elf32_Word type)
{
 int symtab_idx=0;
 int i=0;
 for(i=0; i < elf_hdr_ptr->e_shnum;i++)
   {
     if((sht_ptr->sh_type)==type)
        {
	      if(symtab_idx)
		   {
		      printf("The file appears to have more than one Symbol Table\nOnly one is supported at the moment\n");
		      return symtab_idx;
		    }

		symtab_idx=i;
	}
	
     #ifdef DEBUG
     	printf("%s ",str_t_ptr);
	printf("%s\n",str_t_ptr);
     #endif
	sht_ptr++;
    }
 #ifdef DEBUG2
/* Different message depending on the type of table */
  if(type==SHT_SYMTAB)
   {
     if(symtab_idx)
 	printf("The file has one Symbol Table with index %i\n",symtab_idx);
     else
   	printf("\033[31mThere is no Symbol Table in this file\033[0m\n");
    }
   if(type==SHT_DYNSYM)
   {
     if(symtab_idx)
 	printf("The file has one DYNSYM Table with index %i\n",symtab_idx);
     else
   	printf("\033[31mThere is no DYNSYM Table in this file\033[0m\n");
    }
#endif
	return symtab_idx;
}

/* Allocate memory for Symbol table */
Elf32_Sym* symtab_dyn_memory(Elf32_Shdr* sht_ptr, int idx)
{
/* Align Elf32_Shdr pointer to the SYMTAB structure */
sht_ptr +=idx;
Elf32_Sym* symtab_ptr;
if ((symtab_ptr=malloc(sht_ptr->sh_size))==NULL)
  {
    perror("malloc() failed");
    exit(1);
  }
return symtab_ptr;
}
/* Symbol table printing */
void sym_tab_print(Elf32_Shdr* sht_ptr, Elf32_Sym* symtab_ptr,FILE* fp, int idx, char* str_t_ptr)
{
/* Remember initial position of the Elf32_Shdr pointer */
Elf32_Shdr* sht_ptr_init;
sht_ptr_init = sht_ptr;
char* str_t_ptr_init = str_t_ptr;

/* First we need to find out the String Table associated with 
this Symbol Table */
char* str_tab; 
sht_ptr +=idx;
int str_tab_idx=0;
str_tab_idx = sht_ptr->sh_link;
#ifdef DEBUG2
printf("String tab for Symbol Table has index %i\n",sht_ptr->sh_link);
#endif
sht_ptr = sht_ptr_init + str_tab_idx;
if (( str_tab=malloc(sht_ptr->sh_size))==NULL)
   {
     perror("malloc() failed");
     exit(1);
   }
 /* Read string table into allocated memory */
if ((fseek(fp,sht_ptr->sh_offset,SEEK_SET)))
 {
  perror("fseek() failed");
  exit(1);
  }
int count=0;
if ((count = fread(str_tab,1,sht_ptr->sh_size,fp)) !=sht_ptr->sh_size)
   {
    printf("Error while reading file in pop_str_t\n");
    fclose(fp);
    exit(1);
  }
 /* Print names in the Symbol Table
 First find out number of entries in the Symbol Table*/
 sht_ptr = sht_ptr_init + idx;
 char* str_tab_init = str_tab;
  int num_ent=((sht_ptr->sh_size)/sizeof(Elf32_Sym));
  int i=0;
  int bind=0;
  int type=0;
  printf("\n"); /* Just in case */
  printf("\033[31m%25s\n\033[0m","SYMBOL TABLE");
  symtab_ptr++; /* First entry is reserved, does not need to
  be printed */
  for(i=1;i<num_ent;i++)
  {
   printf("%08x  ",symtab_ptr->st_value);
   bind=ELF32_ST_BIND(symtab_ptr->st_info);
/* Analyze bind */
   switch (bind) {
   	case STB_LOCAL:
		printf("l  ");
		break;
	case STB_GLOBAL:
		printf("g  ");
		break;
	case STB_WEAK:
		printf("w  ");
		break;
	default:
		printf("?  ");
	}
   type=ELF32_ST_TYPE(symtab_ptr->st_info);
/* Analyze type */
   switch (type) {
   	case STT_NOTYPE:
		printf("NOTYPE ");
		break;
	case STT_OBJECT:
		printf("OBJECT ");
		break;
	case STT_FUNC:
		printf("FUNC ");
		break;
	case STT_SECTION:
		printf("SECTION ");
		break;
	case STT_FILE:
		printf("FILE ");
		break;
	default:
		printf("UNKNOWN ");
	}
/* Print name of the relevant section
Requires some work around. First we need 
to check if it's not special value */
	switch (symtab_ptr->st_shndx) {
		case SHN_UNDEF:
			printf("*UND*  ");
			break;
		case SHN_LORESERVE:
			printf("*LORES*  ");
			break;
		case SHN_HIPROC:
			printf("*HIP* ");
			break;
		case SHN_ABS:
			printf("*ABS*  ");
			break;
		case SHN_COMMON:
			printf("*COM*  ");
			break;
		case SHN_HIRESERVE:
			printf("*HIRES*  ");
			break;
		default:
		sht_ptr=sht_ptr_init;
		sht_ptr +=(symtab_ptr->st_shndx);
		str_t_ptr=str_t_ptr_init;
		str_t_ptr +=(sht_ptr->sh_name);
		printf("%s  ",str_t_ptr);
	}
   printf("%08x  ",symtab_ptr->st_size);
   str_tab +=(symtab_ptr->st_name);
   printf("%s\n",str_tab);
   symtab_ptr++;
   str_tab=str_tab_init;
  }
 /* After everything is done we need to free dynamic memory
 occupied by the String Table */
 free(str_tab);
 }
 /* This function is for test purposes. It prints
 Symbol Table entries without string names, everything is in hex */
 void print_symtab_hex(Elf32_Shdr* sht_ptr, Elf32_Sym* symtab_ptr, int idx)
 {
 /* Find out number of entries in the Symbol Table */
 sht_ptr +=idx;
 int num_ent=(sht_ptr->sh_size/sizeof(Elf32_Sym));
 printf("Symbol Table has %i entries\n",num_ent);
 int i=0;
 for (i=0;i<num_ent;i++)
   {
      printf("%i  ",i+1);
      printf("%08x  %08x  %08x  %x  %x  %04x\n",symtab_ptr->st_name, symtab_ptr->st_value, symtab_ptr->st_size, symtab_ptr->st_info, symtab_ptr->st_other, symtab_ptr->st_shndx);
      symtab_ptr++;
   }
 }

/* This function is to populate Symbol Table */
void pop_sym_t(Elf32_Shdr* sht_hdr_ptr, Elf32_Sym* symtab_ptr,FILE* fp, int idx)
{
/* Align file pointer to the beginning of the String Table */
sht_hdr_ptr +=idx;

if ((fseek(fp,sht_hdr_ptr->sh_offset,SEEK_SET)))
 {
  perror("fseek() failed");
  exit(1);
  }

int count=0;
if ((count = fread(symtab_ptr,1,sht_hdr_ptr->sh_size,fp)) !=sht_hdr_ptr->sh_size)
   {
    printf("Error while reading file in pop_str_t\n");
    fclose(fp);
    exit(1);
  }
#ifdef DEBUG
printf("%#010x bytes read into String Table\n",count);
#endif
}

/* Function to allocate dynamic memory for Program
Header Table */

Elf32_Phdr* phdr_dyn_memory(Elf32_Ehdr* elf_header)
  {
    Elf32_Phdr* ph_table_ptr;
  
if(sizeof(Elf32_Phdr)!=elf_header->e_phentsize)
   {
   	printf("sizeof(Elf32_Phdr)!=elf_header.e_phentsize\n");
   	printf("Structure of Program Header Table entry is unknown\n");
	exit(1);
   }
if ((ph_table_ptr= malloc(elf_header->e_phnum*elf_header->e_phentsize))==NULL)
  {
    perror("malloc() failed");
    exit(1);
  }
  return ph_table_ptr;
}

/* Function to  populate dynamic memory for PHT */
void pop_pht(Elf32_Ehdr* elf_header, Elf32_Phdr* pht_ptr, FILE* fp)
{

/* Position pointer of the opened file to the beginning of PHT */
if((fseek(fp,elf_header->e_phoff,SEEK_SET)))
  	{
  	   perror("fseek() failed");
  	   exit(1);
  	}

/* Cycle to read PHT from file into array of structures */
   	
int count=0; 
if ((count = fread(pht_ptr, elf_header->e_phentsize,elf_header->e_phnum,fp)) !=elf_header->e_phnum)
   {
    printf("Error while reading file\n");
    fclose(fp);
    exit(1);
   }
#ifdef DEBUG
printf("%i members in SHT\n",count);
#endif
}

/* Print PHT in readable form */
void print_pht(Elf32_Ehdr* elf_hdr_ptr, Elf32_Phdr* p_hdr_ptr)
{
printf("\nProgram Header:\n");
int i=0;
for (i=0;i<elf_hdr_ptr->e_phnum;i++)
  {
   switch (p_hdr_ptr->p_type) {
   	case PT_NULL:
		printf("NULL  ");
		break;
	case PT_LOAD:
		printf("LOAD  ");
		break;
	case PT_DYNAMIC:
		printf("DYNAMIC  ");
		break;
	case PT_INTERP:
		printf("INTERP  ");
		break;
	case PT_NOTE:
		printf("NOTE  ");
		break;
	case PT_SHLIB:
		printf("SHLIB  ");
		break;
	case PT_PHDR:
		printf("PHDR  ");
		break;
	default:
		printf("UNKNOWN  ");
	}
	
printf("off  %08x ",p_hdr_ptr->p_offset);
printf("vaddr  %08x ",p_hdr_ptr->p_vaddr);
printf("paddr  %08x ",p_hdr_ptr->p_paddr);
printf("align  %x\n\t",p_hdr_ptr->p_align);
printf("filesz  %08x  ",p_hdr_ptr->p_filesz);
printf("memsz  %08x  ",p_hdr_ptr->p_memsz);
/* Analyze flags */
printf("flags ");
if(p_hdr_ptr->p_flags&PF_R)
	printf("r");
		else
		   printf("-");
if(p_hdr_ptr->p_flags&PF_W)
	printf("w");
		else
		   printf("-");
if(p_hdr_ptr->p_flags&PF_X)
	printf("x\n");
		else
		   printf("-\n");
p_hdr_ptr++;
  }
}

/* Function to check if the PHT has Dynamic Section.
If PHT section has Dynamic Section, this function returns
its index */
int if_dyn_pht(Elf32_Ehdr* elf_hdr_ptr, Elf32_Phdr* pht_ptr)
{
int dyn_idx=0xffffffff;
 int i=0;
 for(i=0; i < elf_hdr_ptr->e_phnum;i++)
   {
     if((pht_ptr->p_type)==PT_DYNAMIC)
        {
	      if(dyn_idx!=0xffffffff)
		   {
		      printf("The file appears to have more than one Dynamic Table in the PHT\nOnly one is supported at the moment\n");
		      return dyn_idx;
		    }

		dyn_idx=i;
	}
	
	pht_ptr++;
    }
	return dyn_idx;
}

/* Function to allocate memory for DYNAMIC segment */
Elf32_Dyn* dyn_dyn_memory(Elf32_Phdr* pht_ptr,int idx)
{

 Elf32_Dyn* dyn_ptr;
/* Align pht_ptr to DYNAMIC structure */
pht_ptr +=idx;
  
if ((dyn_ptr= malloc(pht_ptr->p_filesz))==NULL)
  {
    perror("malloc() failed");
    exit(1);
  }
  return dyn_ptr;
}

/* Populate allocated for DYNAMIC segment memory */
void pop_dynamic(Elf32_Dyn* dyn_ptr, Elf32_Phdr* pht_ptr, FILE* fp, int idx)
{
/* Set pht_ptr to DYNAMIC structure */
pht_ptr +=idx;
/* Position pointer of the opened file to the beginning of DYNAMIC
section */
if((fseek(fp,pht_ptr->p_offset,SEEK_SET)))
  	{
  	   perror("fseek() failed");
  	   exit(1);
  	}

/* Cycle to read DYNAMIC segment from file into array of structures */
   	
int count=0; 
if ((count = fread(dyn_ptr,1,pht_ptr->p_filesz,fp)) !=pht_ptr->p_filesz)
   {
    printf("Error while reading file in pop_dynamic()\n");
    fclose(fp);
    exit(1);
   }
#ifdef DEBUG
printf("%i bytes in DYNAMIC\n",count);
#endif
}

/* Print content of the DYNAMIC segment */
void print_dynamic(Elf32_Phdr* pht_ptr, Elf32_Dyn* dyn_ptr, FILE* fp, univ_struct* un_struct)
{
Elf32_Dyn* dyn_ptr_init = dyn_ptr;
Elf32_Word strtab_siz=0;
char* strtab_ptr;
char* strtab_ptr_init;
/* Find out the size of the String table (assume it exists).
If the table doesn't exist, it will just cycle through in idle
mode */
while(dyn_ptr->d_tag!=DT_NULL)
   {
   	if(dyn_ptr->d_tag==DT_STRSZ)
	   {
	   	strtab_siz=dyn_ptr->d_un.d_val;
		break;
	   }
      dyn_ptr++;
    }
 dyn_ptr=dyn_ptr_init; /* Set pointer to the beginning again */
 
/* We need to find and allocate memory for STRING TABLE, if it
exists */
while(dyn_ptr->d_tag!=DT_NULL)
   {
   	if(dyn_ptr->d_tag==DT_STRTAB)
	  {
	     if((strtab_ptr=malloc(strtab_siz))==NULL)
	        {
		  perror("malloc() in print_dynamic() failed");
		  exit(1);
		}
	     if((fseek(fp,un_struct->dynstr_offset,SEEK_SET)))
  		{
  	   	   perror("fseek() failed");
  	   	   exit(1);
  		}

/* Cycle to read DYNAMIC segment from file into array of structures */
   	
		int count=0; 
	     if ((count = fread(strtab_ptr,1,strtab_siz,fp)) !=strtab_siz)
   		{
    		   printf("Error while reading file in print_dynamic()\n");
		   printf("%i bytes read, while %i should be read\n",count,strtab_siz);
    		   fclose(fp);
    		   exit(1);
   		}
		break;
	   } /* End of if */
   	dyn_ptr++;
   }
/* Realign dyn_ptr again */
dyn_ptr=dyn_ptr_init;

/* Last structure in this array must be of type DT_NULL.
This is how we find out if we've read all we need to read */
printf("\n\tDynamic Section:\n");
/* Before we start printing remember position of the beginning of the
string tab */
strtab_ptr_init=strtab_ptr;
while(dyn_ptr->d_tag!=DT_NULL)
  {
/* Parsing of every member requires individual approach */
	switch (dyn_ptr->d_tag) {
	   	case DT_NULL:
		/* This should never be the case */
			printf("NULL  ");
			break;
		case DT_NEEDED:
			printf("%-20s","NEEDED");
			strtab_ptr +=dyn_ptr->d_un.d_val;
			printf("%s",strtab_ptr);
			strtab_ptr=strtab_ptr_init;
			break;
		case DT_PLTRELSZ:
			printf("%-20s","PLTRELSZ");
			printf("%08x",dyn_ptr->d_un.d_val);
			break;
		case DT_PLTGOT:
			printf("%-20s","PLTGOT");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_HASH:
			printf("%-20s","HASH");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_STRTAB:
			printf("%-20s","STRTAB");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_SYMTAB:
			printf("%-20s","SYMTAB");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_RELA:
			printf("%-20s","RELA");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_RELASZ:
			printf("%-20s","RELASZ");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_RELAENT:
			printf("%-20s","RELAENT");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_STRSZ:
			printf("%-20s","STRSZ");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_SYMENT:
			printf("%-20s","SYMENT");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_INIT:
			printf("%-20s","INIT");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_FINI:
			printf("%-20s","FINI");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_SONAME:
			printf("%-20s","SONAME");
			strtab_ptr +=dyn_ptr->d_un.d_val;
			printf("%s",strtab_ptr);
			strtab_ptr=strtab_ptr_init;
			break;
		case DT_RPATH:
			printf("%-20s","RPATH");
			strtab_ptr +=dyn_ptr->d_un.d_val;
			printf("%s",strtab_ptr);
			strtab_ptr=strtab_ptr_init;
			break;
		case DT_SYMBOLIC:
			printf("%-20s","SYMBOLIC");
			printf("Value is ignored");
			break;
		case DT_REL:
			printf("%-20s","REL");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_RELSZ:
			printf("%-20s","RELSZ");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_RELENT:
			printf("%-20s","RELENT");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_PLTREL:
			printf("%-20s","PLTREL");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_DEBUG:
			printf("%-20s","DEBUG");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_TEXTREL:
			printf("%-20s","TEXTREL");
			printf("Value is ignored"); 
			break;
		case DT_JMPREL:
			printf("%-20s","JMPREL");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		default:
			printf("%-20s","UNKNOWN");
			printf("%08x",dyn_ptr->d_un.d_ptr);
		}
	printf("\n");
	dyn_ptr++;
   }
/* After everything is done we need to free dynamically
allocated memory */
free(strtab_ptr);
}

/* This function is for debug purposes. It prints content of the
DYNAMIC segment with names in symbol representation but all values in
hex, no matter if string or something else is expected */
void print_dynamic_hex(Elf32_Phdr* pht_ptr, Elf32_Dyn* dyn_ptr)
{
printf("\n");
while(dyn_ptr->d_tag!=DT_NULL)
  {
/* Parsing of every member requires individual approach */
	switch (dyn_ptr->d_tag) {
	   	case DT_NULL:
		/* This should never be the case */
			printf("NULL  ");
			break;
		case DT_NEEDED:
			printf("NEEDED  ");
			printf("%08x",dyn_ptr->d_un.d_val);
			break;
		case DT_PLTRELSZ:
			printf("PLTRELSZ  ");
			break;
		case DT_PLTGOT:
			printf("PLTGOT  ");
			break;
		case DT_HASH:
			printf("HASH  ");
			break;
		case DT_STRTAB:
			printf("STRTAB  ");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_SYMTAB:
			printf("SYMTAB  ");
			printf("%08x",dyn_ptr->d_un.d_ptr);
			break;
		case DT_RELA:
			printf("RELA  ");
			break;
		case DT_RELASZ:
			printf("RELASZ  ");
			break;
		case DT_RELAENT:
			printf("RELAENT  ");
			break;
		case DT_STRSZ:
			printf("STRSZ  ");
			break;
		case DT_SYMENT:
			printf("SYMENT  ");
			break;
		case DT_INIT:
			printf("INIT  ");
			break;
		case DT_FINI:
			printf("FINI  ");
			break;
		case DT_SONAME:
			printf("SONAME  ");
			break;
		case DT_RPATH:
			printf("RPATH  ");
			break;
		case DT_SYMBOLIC:
			printf("SYMBOLIC  ");
			break;
		case DT_REL:
			printf("REL  ");
			break;
		case DT_RELSZ:
			printf("RELSZ  ");
			break;
		case DT_RELENT:
			printf("RELENT  ");
			break;
		case DT_PLTREL:
			printf("PLTREL  ");
			break;
		case DT_DEBUG:
			printf("DEBUG  ");
			break;
		case DT_TEXTREL:
			printf("TEXTREL  ");
			break;
		case DT_JMPREL:
			printf("JMPREL  ");
			break;
		default:
			printf("UNKNOWN  ");
		}
	printf("\n");
	dyn_ptr++;
   }
}
