//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// FIFO module. Size is 512 bytes
//
module fifo (
  clk_i, rst_i, we_i, rd_i, data_i,
  ffull_o, fempty_o, data_o, fifo_count_o
  );
  
  input clk_i;                  // Input clock signal
  input rst_i;                  // Input reset signal
  input we_i;                   // Write enable signal
  input rd_i;                   // Read FIFO
  input [7:0] data_i;           // Input data from CPU
  output ffull_o;               // Flag that FIFO is full
  output fempty_o;              // Flag that FIFO is empty
  output [7:0] data_o;          // Output data read from FIFO
  output [8:0] fifo_count_o;          // Number of bytes in FIFO
 
  reg [7:0] data_o;
 // Internal wires and registers 
  reg [7:0] fifo [0:511];       // FIFO - 512 bytes
  reg [8:0] adrr;               // Read address
  reg [8:0] adrw;               // Write address
  reg [8:0] fifo_count;         // Count of words in FIFO
  reg fifo_full;               // Flag that FIFO is full
  reg fifo_empty;              // Flag that FIFO is empty
  
  assign ffull_o = fifo_full;
  assign fempty_o = fifo_empty;
  assign fifo_count_o = fifo_count;
   
  always @ (posedge clk_i)          // Write FIFO logic
    begin
      if (!rst_i)       // Reset condition
        begin
          fifo_full <= 1'b0;
          adrw <= 9'h0;
        end
      else if (we_i)
        begin
          fifo[adrw] <= data_i;
          adrw <= adrw + 1'b1;
          if ((adrw + 1'b1) === adrr)           // Check if FIFO is full
            fifo_full <= 1'b1;							  
          else
            fifo_full <= 1'b0;
        end
      else if (adrw != adrr)
        fifo_full <= 1'b0;             // If not during write cycle read and write addresses not equal FIFO is not full	   
    end
    
  always @ (posedge clk_i)          // Read FIFO logic
    begin
      if (!rst_i)
        begin
          fifo_empty <= 1'b1;      // FIFO is empty upon reset
          adrr <= 9'h0;
        end
      else if (rd_i)
        begin
          data_o <= fifo[adrr];
          adrr <= adrr + 1'b1;
          if ((adrr + 1'b1 === adrw))
            fifo_empty <= 1'b1;        
        end  
      else if (adrw !== adrr)
        fifo_empty <= 1'b0;          // If not during read cycle read and write addresses not equal FIFO is not empty
    end
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          fifo_count <= 9'h0;
        end
      else
        begin
          if (rd_i && ~we_i && ~fifo_empty)
            fifo_count <= fifo_count - 1'b1;
          else if (~rd_i && we_i)
            fifo_count <= fifo_count + 1'b1;
        end
    end
      
endmodule