//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Hazard detection unit
module hazard (
  clk_i, rst_i, stallm_i, load_i, rse_i, rte_i, rlm_i, stallmdu_i, sys_eslip_i,
  stalli_o, stalle_o, stallm_o, stalla_o, stallw_o, slipe_o, stall_load_o
  );
  
  input clk_i;            // Input clock
  input rst_i;            // Input reset signal
  input stallm_i;         // Input from the M stage that stall is required
  input load_i;           // Signal that instruction is of load type
  input [4:0] rse_i;      // Address of the rs register from E stage
  input [4:0] rte_i;      // Address of the rt register from E stage
  input [4:0] rlm_i;      // Address of the resister to be loaded from the M stage
  input stallmdu_i;       // Input signal to stall from MDU unit
  input sys_eslip_i;      // Input signal from the CP0 to slip E stage
  output stalli_o;        // Output to stall I stage
  output stalle_o;        // Output to stall E stage
  output stallm_o;        // Output to stall M stage
  output stalla_o;        // Output to stall A stage
  output stallw_o;        // Output to stall W stage
  output slipe_o;         // Output to slip E stage
  output stall_load_o;    // Output signal that LOAD stall has happened
  
  // Internal regs and wires
  reg loadi_buf;          // Buffer to delay for 0.5 clk
  reg [4:0] rsei_buf;     // Buffer to delay for 0.5 clk
  reg [4:0] rtei_buf;     // Buffer to delay for 0.5 clk
  reg [4:0] rlmi_buf;     // Buffer to delay for 0.5 clk
  
  assign stall_load_o = loadi_buf && (&(rsei_buf^~rlmi_buf) || &(rtei_buf^~rlmi_buf));        // Stall condition caused by load instruction
  
  
  wire [1:0] stall_logic = {stallm_i, (stall_load_o|stallmdu_i)};
  wire [1:0] slip_logic = {stallm_i, (stall_load_o|stallmdu_i|sys_eslip_i)};
 
  
  // Stall logic 
  // All stall signals must be asserted on the posedge of clock
  assign {stalli_o,stalle_o,stallm_o,stalla_o,stallw_o} = (stall_logic === 2'b10) ? 5'b11100:
                                                          (stall_logic === 2'b01) ? 5'b10000:
                                                          (stall_logic === 2'b11) ? 5'b11100:
                                                                                    5'b00000;
                                                                                
  // Slip logic
  assign slipe_o = (slip_logic === 2'b01) ? 1'b1:
                                            1'b0;
                                                                                
  always @ (posedge clk_i)
    begin
      loadi_buf <= load_i;            // Delay 0.5 clk
      rsei_buf <= rse_i;              // Delay 0.5 clk
      rtei_buf <= rte_i;              // Delay 0.5 clk
      rlmi_buf <= rlm_i;              // Delay 0.5 clk
    end

endmodule