//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//		 
// Module for RAM memory
//
module memory (
  clk_i, we_i, data_i, be_i, adr_i,
  data_o
);
  input clk_i;            // Input clock
  input we_i;             // Write enable signal
  input [31:0] data_i;    // Input data
  input [3:0] be_i;       // Byte mask
			
  input [12:0] adr_i;     // Input address
  output [31:0] data_o;   // Output data
  
  reg [31:0] data_o;
  wire wei_1 = we_i & be_i[0]; 
  wire wei_2 = we_i & be_i[1];
  wire wei_3 = we_i & be_i[2];
  wire wei_4 = we_i & be_i[3];
  
  // Internal registers and wires
   reg [7:0] data1 [0:8191];        // Array of first byte of data, LSB
   reg [7:0] data2 [0:8191];        // Array of second byte of data
   reg [7:0] data3 [0:8191];        // Array of third byte of data
   reg [7:0] data4 [0:8191];        // Array of fourth byte of data, MSB
 
/* 
	initial
	  begin
	    $readmemh("memory_b0.list",data1);
	    $readmemh("memory_b1.list",data2);
	    $readmemh("memory_b2.list",data3);
	    $readmemh("memory_b3.list",data4);
	  end				
		*/
  always @ (posedge clk_i)       
    begin
      if (wei_1)
        data1[adr_i] <= data_i[7:0];
      if (wei_2)
        data2[adr_i] <= data_i[15:8];
      if (wei_3)
        data3[adr_i] <= data_i[23:16];
      if (wei_4)
        data4[adr_i] <= data_i[31:24];
      data_o <= {data4[adr_i],data3[adr_i],data2[adr_i],data1[adr_i]};
    end
endmodule