//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Regfile
module regfile (
  clk_i, rst_i, rd_i, rda_i, rsa_i, rta_i, wen_i,
  rdms_i, rdas_i, uart_txrdy_i,
  rs_o, rt_o, uart_wr_o, uart_data_o
  
  );
  input clk_i;          // Clock input
  input rst_i;          // Input reset signal
  input [31:0] rd_i;    // Destination register from W stage 
  input [4:0] rda_i;    // Address for the destination register
  input [4:0] rsa_i;    // Address for the source register
  input [4:0] rta_i;    // Address for the source register
  input wen_i;          // Write enable signal
  input [31:0] rdms_i;  // Input rd register from M stage
  input [31:0] rdas_i;  // Input rd register from A stage
  input uart_txrdy_i;   // Input from UART
  output [31:0] rs_o;   // Source register for ALU 
  output [31:0] rt_o;   // Source register for ALU
  output uart_wr_o;     // Output to UART
  output [7:0] uart_data_o; // Output to UART
  
  
  reg [31:0] reg_bank[0:31];   // Bank of general purpose registers
  
  // Outputs of the register file including bypass stage
  assign rs_o = ((wen_i === 1'b1)&&(rsa_i === rda_i)) ? rd_i : reg_bank[rsa_i];
  assign rt_o = ((wen_i === 1'b1)&&(rta_i === rda_i)) ? rd_i : reg_bank[rta_i];
 	
   // Test internal registers and wires
   reg [3:0] sw_r = 4'h0;
   reg [3:0] strap_r = 4'h0;
   reg [9:0] count_r = 10'h81;
  // Instantiate modules  
  test_m uart_link (
    .clk_i                (clk_i),
    .rst_i                (rst_i),
    .sdata1_i             (reg_bank[5'h3]),
    .sdata2_i             (reg_bank[5'h5]),
    .sw_i                 (sw_r),
    .strap_i              (strap_r),
    .count_i              (count_r),
    .uart_txrdy_i         (uart_txrdy_i),
    .start_i              (rst_i),
    .uart_wr_o            (uart_wr_o),
    .uart_data_o          (uart_data_o)
  );                                            
  
always @ (posedge clk_i)
  begin			
    if (!rst_i)
      begin
        reg_bank[4'h0] <= 32'h0;            // Initialize R0 to zero value  
      end
    else if (wen_i && (rda_i !== 4'h0))          // Do not write data if destination register is R0  						  
        reg_bank[rda_i] <= rd_i;          // Update register
        
  end
endmodule
